/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2009, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert command line text output into XML formatted output     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "schema.h"

/* function templates */
void load_dtd();
void create_element_record_dtd();


void load_dtd()
{
  /* Read the DTD file and parse it into relevant structure members.
   * There is plenty of scope for improvement in this subroutine as
   * parsing routines are notoriously messy in C.
   *
   * Any comments that may be encountered are simply filtered out 
   * of the processing.
   *
   * The full DTD functionality is not currently supported.
   */
  long i, lead_trim;
  unsigned char c;

  flag_comments        = FALSE;
  flag_char            = FALSE;
  schema_parent.nelements = 0;
  schema_parent.nrecords  = 0;
  schema_parent.nlevels   = 0;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: load_dtd\n");

  /* initialise field handling variable */
  field_value = (unsigned char **) malloc(DEFAULT_NFIELDS * sizeof(unsigned char *));
  if (field_value == NULL) out_of_mem();

  utf8_getchar_init();
  if ((schema_fd = open((char *) schema_filename, O_RDONLY)) > 0)
  {
    while (!utf8_flag_EOF)
    {
      flag_char            = FALSE;
      input_record.length  = 0;
      input_record.nfields = 0;

      /* read line into array */
      while ((c = utf8_getchar(schema_fd)) == c && !utf8_flag_EOF && c != '>')
      {
        /* ignore carriage returns and new lines */
        if (c == '\n' || c == '\r') 
          continue;
          
        input_record.line[input_record.length++] = c;

        /* mem checks */
        if (input_record.length > memseg_input_record)
          memadd_input_record();
      }

      /* don't process EOF */
      if (utf8_flag_EOF)
        break;

      /* put back the loop terminator */
      if (c == '>')
        input_record.line[input_record.length++] = c;

      /* remove leading trim */
      lead_trim = 0;
      while(lead_trim < input_record.length && isspace(input_record.line[lead_trim]))
        lead_trim++;
      if (lead_trim)
      {
        input_record.length -= lead_trim;
        for (i = 0 ; i < input_record.length ; i++)
          input_record.line[i] = input_record.line[i + lead_trim];
      }
      input_record.line[input_record.length] = '\0';

      /* check if args also specified inside schema file */
      schema_args();

      /* purge comments and remove blank lines */
      if ((input_record.length = purge_comments_blanklines(input_record.line, input_record.length)) == 0)
        continue;

      /* ignore lines beginning with <? like <?xml because some DTD writers wrongfully include it */
      if (input_record.length > 1 && xstrncmp((unsigned char *) "<?", input_record.line, 2) == 0)
      {
        if (sw_debug)
          fprintf(stderr, "DEBUG: DTD: Ignoring unsupported schema line: %s\n", input_record.line);
        continue;
      }

      /* ignore lines beginning with <!DOCTYPE because some DTD writers wrongfully include it */
      if (input_record.length > 8 && xstrncmp((unsigned char *) "<!DOCTYPE", input_record.line, 9) == 0)
      {
        if (sw_debug)
          fprintf(stderr, "DEBUG: DTD: Ignoring unsupported schema line: %s\n", input_record.line);
        continue;
      }

      /* break line into fields */
      for (i = 0 ; i < input_record.length ; i++)
      {
        if (! isspace(input_record.line[i])) 
        {
          if (!flag_char) 
            field_value[input_record.nfields++] = &input_record.line[i];
          flag_char = TRUE;
        }
        else
        {
          if (flag_char) 
            input_record.line[i] = '\0';
          flag_char = FALSE;
        } 

        /* mem checks */
        if (input_record.nfields > memseg_nfields)
        {
          memadd_nfields();
          field_value = (unsigned char **) realloc(field_value, (memseg_nfields + 10) * sizeof(unsigned char *));
          if (field_value == NULL) out_of_mem();
        }
      }

      /* Load up <!ELEMENT lines */
      if (input_record.nfields > 1 && xstrcmp(field_value[0], (unsigned char *) "<!ELEMENT") == 0)
        create_element_record_dtd();

    }
    close(schema_fd);

    create_elements();
    if (err_status == 0) build_element_hierachy();
    if (err_status == 0) helper_calc_tree_paths();
  }
  else
  {
    fprintf(stderr, "%s: ERROR: Cannot open DTD file %s", PROGNAME, schema_filename);
    err_status = -2;
  }

  /* cleanup */
  free(field_value); /* pointer array */

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: load_dtd\n");
}


void create_element_record_dtd()
{
  /* Populate the schema_record and schema_field structures */
  long i, j, k, l;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: create_element_record\n");

  /* Filter out elements that xmlfy currently does not support */
  for (i = 0 ; i < input_record.nfields ; i++)
  {
    /* don't process element records containing contentspec of (#CDATA) or (#PCDATA) */
    if (xstrncmp(field_value[i], (unsigned char *) "(#CDATA)", 8) == 0 || xstrncmp(field_value[i], (unsigned char *) "(#PCDATA)", 9) == 0)
    {
      if (sw_debug)
      {
        fprintf(stderr, "DEBUG: DTD: found \"(#CDATA)|(#PCDATA)\" element definition.\n");
        fprintf(stderr, "DEBUG: end function: create_element_record\n");
      }
      return;
    }
    /* don't process element records containing contentspec of EMPTY or ANY */
    if (xstrncmp(field_value[i], (unsigned char *) "EMPTY", 5) == 0 || xstrncmp(field_value[i], (unsigned char *) "ANY", 3) == 0)
    {
      if (sw_debug)
      {
        fprintf(stderr, "DEBUG: DTD: found \"EMPTY|ANY\" element definition.\n");
        fprintf(stderr, "DEBUG: end function: create_element_record\n");
      }
      return;
    }
  }

  if (schema_parent.nrecords == 0)
  {
    schema_record = (struct_schema_record *) malloc(sizeof(struct_schema_record));
    if (schema_record == NULL) out_of_mem();
  }
  else
  {
    schema_record = (struct_schema_record *) realloc(schema_record, (schema_parent.nrecords + 1) * sizeof(struct_schema_record));
    if (schema_record == NULL) out_of_mem();
  }

  schema_record[schema_parent.nrecords].name = (unsigned char *) malloc((xstrlen(field_value[1]) + 1) * sizeof(unsigned char));
  if (schema_record[schema_parent.nrecords].name == NULL) out_of_mem();
  xstrcpy(schema_record[schema_parent.nrecords].name, field_value[1]);
  schema_record[schema_parent.nrecords].record_count = 0;
  schema_record[schema_parent.nrecords].nfields = 0;

  schema_record[schema_parent.nrecords].field = (struct_schema_field *) malloc((input_record.nfields - 2) * sizeof(struct_schema_field));
  if (schema_record[schema_parent.nrecords].field == NULL) out_of_mem();
  /* set kl markers */
  j = 0;
  while (input_record.line[j] != '(' && j < input_record.length) j++; /* move to first open bracket */

  while (j < input_record.length)
  {
    while (!is_legal_xml_element_name_char(input_record.line[j]) && j < input_record.length) j++; /* move past all noise chars */
    k = j; /* mark the start of field */
    while (is_legal_xml_element_name_char(input_record.line[j]) && j < input_record.length) j++; /* mark the end of field */
    if (j - k > 0)
    {
      schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name = (unsigned char *) malloc((j - k + 1) * sizeof(unsigned char));
      if (schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name == NULL) out_of_mem();
      for (l = 0 ; l < j - k ; l++)
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name[l] = input_record.line[k + l];
      schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name[l] = '\0';

      /* field match_control */
      schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].match_control = match_control_token(input_record.line[j]); 

      /* set initial field type until further examination */
      schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].type = FIELD_CDATA; 
      /* set initial field count */
      schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].field_count = 0; 

      schema_record[schema_parent.nrecords].nfields += 1;
    }
    j++;
  }

  if (sw_debug)
  {
    fprintf(stderr, "DEBUG: Schema: record(%d) = %s\n", schema_parent.nrecords, schema_record[schema_parent.nrecords].name);
    for (j = 0 ; j < schema_record[schema_parent.nrecords].nfields ; j++)
      fprintf(stderr, "DEBUG: Schema:   field(%ld) = %s match_control=%d\n", j, schema_record[schema_parent.nrecords].field[j].name, schema_record[schema_parent.nrecords].field[j].match_control);
  }

  schema_parent.nrecords++;

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: create_element_record\n");
}
