/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2009, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert command line text output into XML formatted output     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "schema.h"


/* function templates */
void load_rnc();
void create_element_record_rnc();


void load_rnc()
{
  /* Read the RELAX NG Compact file and parse it into relevant structure members.
   * There is plenty of scope for improvement in this subroutine as
   * parsing routines are notoriously messy in C.
   *
   * Any comments that may be encountered are simply filtered out 
   * of the processing.
   *
   * The full RELAX NG Compact functionality is not currently supported.
   */
  long i, lead_trim;
  unsigned char c;

  flag_comments           = FALSE;
  flag_char               = FALSE;
  schema_parent.nelements = 0;
  schema_parent.nrecords  = 0;
  schema_parent.nlevels   = 0;
  rnc_start_element_name  = NULL;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: load_rnc\n");

  /* initialise field handling variable */
  field_value = (unsigned char **) malloc(DEFAULT_NFIELDS * sizeof(unsigned char *));
  if (field_value == NULL) out_of_mem();

  utf8_getchar_init();
  if ((schema_fd = open((char *) schema_filename, O_RDONLY)) > 0)
  {
    while (!utf8_flag_EOF)
    {
      flag_char            = FALSE;
      input_record.length  = 0;
      input_record.nfields = 0;

      /* read whole line into array */
      while ((c = utf8_getchar(schema_fd)) == c && !utf8_flag_EOF && c != '\n')
      {
        /* ignore carriage returns */
        if (c == '\r') 
          continue;

        input_record.line[input_record.length++] = c;

        /* mem checks */
        if (input_record.length > memseg_input_record)
          memadd_input_record();
      }

      /* don't process EOF */
      if (utf8_flag_EOF)
        break;

      /* remove leading trim */
      lead_trim = 0;
      while(lead_trim < input_record.length && isspace(input_record.line[lead_trim]))
        lead_trim++;
      if (lead_trim)
      {
        input_record.length -= lead_trim;
        for (i = 0 ; i < input_record.length ; i++)
          input_record.line[i] = input_record.line[i + lead_trim];
      }
      input_record.line[input_record.length] = '\0';

      /* purge blank lines */
      if (input_record.length == 0)
        continue;

      /* check if args also specified inside schema file */
      schema_args();

      /* purge comment lines */
      if (input_record.length > 1 && xstrncmp((unsigned char *) "##", input_record.line, 2) == 0)
      {
        if (sw_debug)
          fprintf(stderr, "DEBUG: RNC: Ignoring comment line: %s\n", input_record.line);
        continue;
      }

      /* break line into fields */
      for (i = 0 ; i < input_record.length ; i++)
      {
        if (! isspace(input_record.line[i])) 
        {
          if (!flag_char) 
            field_value[input_record.nfields++] = &input_record.line[i];
          flag_char = TRUE;
        }
        else
        {
          if (flag_char) 
            input_record.line[i] = '\0';
          flag_char = FALSE;
        } 

        /* mem checks */
        if (input_record.nfields > memseg_nfields)
        {
          memadd_nfields();
          field_value = (unsigned char **) realloc(field_value, (memseg_nfields + 10) * sizeof(unsigned char *));
          if (field_value == NULL) out_of_mem();
        }
      }

      if (input_record.nfields == 3 && xstrcmp(field_value[0], (unsigned char *) "start") == 0
          && (xstrcmp(field_value[1], (unsigned char *) "=") == 0))
      {
        /* Load up parent line */
        rnc_start_element_name = (unsigned char *) malloc((xstrlen(field_value[2]) + 1) * sizeof(unsigned char));
        if (rnc_start_element_name == NULL) out_of_mem();
        xstrcpy(rnc_start_element_name, field_value[2]);
      }
      else if (input_record.nfields > 2 && xstrcmp(field_value[2], (unsigned char *) "element") == 0
          && (xstrcmp(field_value[1], (unsigned char *) "=") == 0 || xstrcmp(field_value[1], (unsigned char *) "|=") == 0))
      {
        /* Load up element lines */
        create_element_record_rnc();
      }
      else if (sw_debug)
      {
        fprintf(stderr, "DEBUG: RNC: Ignoring unsupported schema line: ");
        for (i = 0 ; i < input_record.nfields ; i++)
          fprintf(stderr, "%s ", field_value[i]);
        fprintf(stderr, "\n");
      }

    }
    close(schema_fd);

    if (rnc_start_element_name != NULL)
    {
      create_elements();
      if (err_status == 0) build_element_hierachy();
      if (err_status == 0) helper_calc_tree_paths();
    }
    else
    {
      fprintf(stderr, "%s: ERROR: No start element defined in RNC file %s", PROGNAME, schema_filename);
      err_status = -2;
    }
  }
  else
  {
    fprintf(stderr, "%s: ERROR: Cannot open RNC file %s", PROGNAME, schema_filename);
    err_status = -2;
  }

  /* cleanup */
  free(field_value); /* pointer array */
  if (rnc_start_element_name != NULL)
    free(rnc_start_element_name); 

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: load_rnc\n");
}


void create_element_record_rnc()
{
  /* Populate the schema_record and schema_field structures */
  long j, k, l;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: create_element_record\n");

  /* Filter out elements that xmlfy currently does not support */
  /* don't process element records containing contentspec of { text } */
  if (input_record.nfields > 5 && field_value[4][0] == '{' && xstrncmp(field_value[5], (unsigned char *) "text", 4) == 0 && field_value[6][0] == '}')
  {
    if (sw_debug)
    {
      fprintf(stderr, "DEBUG: RNC: found \"{ text }\" element definition.\n");
      fprintf(stderr, "DEBUG: end function: create_element_record\n");
    }
    return;
  }

  if (schema_parent.nrecords == 0)
  {
    schema_record = (struct_schema_record *) malloc(sizeof(struct_schema_record));
    if (schema_record == NULL) out_of_mem();
  }
  else
  {
    schema_record = (struct_schema_record *) realloc(schema_record, (schema_parent.nrecords + 1) * sizeof(struct_schema_record));
    if (schema_record == NULL) out_of_mem();
  }

  schema_record[schema_parent.nrecords].name = (unsigned char *) malloc((xstrlen(field_value[3]) + 1) * sizeof(unsigned char));
  if (schema_record[schema_parent.nrecords].name == NULL) out_of_mem();
  xstrcpy(schema_record[schema_parent.nrecords].name, field_value[3]);
  schema_record[schema_parent.nrecords].record_count = 0;
  schema_record[schema_parent.nrecords].nfields = 0;

  /* 5 is derived from <name>(0) =(1) element(2) <name>(3) {(4) ... }(5) */
  if (input_record.nfields > 5)
  {
    schema_record[schema_parent.nrecords].field = (struct_schema_field *) malloc((input_record.nfields - 5) * sizeof(struct_schema_field));
    if (schema_record[schema_parent.nrecords].field == NULL) out_of_mem();
    /* set kl markers */
    j = 0;
    while (input_record.line[j] != '{' && j < input_record.length) j++; /* move to first open bracket */
  
    while (j < input_record.length)
    {
      while (!is_legal_xml_element_name_char(input_record.line[j]) && j < input_record.length) j++; /* move past all noise chars */
      k = j; /* mark the start of field */
      while (is_legal_xml_element_name_char(input_record.line[j]) && j < input_record.length) j++; /* mark the end of field */
      if (j - k > 0)
      {
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name = (unsigned char *) malloc((j - k + 1) * sizeof(unsigned char));
        if (schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name == NULL) out_of_mem();
        for (l = 0 ; l < j - k ; l++)
          schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name[l] = input_record.line[k + l];
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name[l] = '\0';
  
        /* field match_control */
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].match_control = match_control_token(input_record.line[j]); 
  
        /* set initial field type until further examination */
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].type = FIELD_CDATA; 
        /* set initial field count */
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].field_count = 0; 
  
        schema_record[schema_parent.nrecords].nfields += 1;
      }
      j++;
    }
  
    if (sw_debug)
    {
      fprintf(stderr, "DEBUG: RNC: record(%d) = %s\n", schema_parent.nrecords, schema_record[schema_parent.nrecords].name);
      for (j = 0 ; j < schema_record[schema_parent.nrecords].nfields ; j++)
        fprintf(stderr, "DEBUG: RNC:   field(%ld) = %s match_control=%d\n", j, schema_record[schema_parent.nrecords].field[j].name, schema_record[schema_parent.nrecords].field[j].match_control);
    }

    schema_parent.nrecords++;
  }

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: create_element_record\n");
}
