/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2009, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert command line text output into XML formatted output     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "unxml.h"

int flag_comments;
long i_pos, t_open, t_close;

/* function templates */
int  is_tag();
void insert_delimiter(int level);
void unescape_chars();


void unxml()
{
  /* Do the opposite of xmlfy and remove all that XML bracket racket.
   * Full XML detoxification is not fully supported as this routine
   * focuses primarily on undoing what xmlfy normally does.
   * 
   * In hindsight, and in staying with Un*x tradition, I should have
   * written two tools and called them intox and detox respectively.
   */
  int level;
  unsigned char c;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: unxml\n");

  /* initialise */
  flag_comments = FALSE;
  level = -1;

  utf8_getchar_init();
  while (!utf8_flag_EOF)
  {
    input_record.length = 0;

    while ((c = utf8_getchar(STDIN_FD)) == c && !utf8_flag_EOF && c != '>')
    {
      input_record.line[input_record.length++] = c;

      /* mem checks */
      if (input_record.length > memseg_input_record)
        memadd_input_record();
    }

    /* don't process EOF */
    if (utf8_flag_EOF)
      break;

    /* put back the loop terminator */
    if (c == '>')
    {
      t_close = input_record.length;
      input_record.line[input_record.length++] = c;
    }

    /* purge comments and remove blank lines */
    if ((input_record.length = purge_comments_blanklines(input_record.line, input_record.length)) == 0)
      continue;

    /* remove XML */
    output_record_length = 0;
    i_pos = 0;
    while (i_pos < input_record.length)
    {
      switch(is_tag())
      {
        case TAG_OPEN:
               /* move to new level */
               level++;
               /* only print delimiter if more than one field */
               if (level >= 0 && level < nlevels)
               {
                 if (field_cntr[level] > 0)
                   insert_delimiter(level);
                 field_cntr[level]++;
               }
               else
                 insert_delimiter(level);
               break;

        case TAG_CLOSE:
               /* clear field counters of next higher level */
               if (level + 1 >= 0 && level + 1 < nlevels)
                 field_cntr[level + 1] = 0;
               /* move to new level */
               level--;
               break;

        case TAG_OPEN_CLOSE:
               break;

        case TAG_PROLOGUE:
               break;

        case TAG_DOCTYPE:
               break;

        default:
               if (level >= 0)
                 output_record[output_record_length++] = input_record.line[i_pos++];
               else
                 i_pos++;

               /* mem checks */
               if (output_record_length > memseg_output_record)
                 memadd_output_record(output_record_length);
      }
    }
    unescape_chars();
    utf8_print(STDOUT_FD, output_record, output_record_length);
  }

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: unxml\n");
}


int is_tag()
{
  /* Determine the tag type */
  int ret;
  long i, lead_trim;

  ret = FALSE;
  i   = 0;

  if (input_record.line[i_pos] == '<')
  {
    t_open = i_pos;

    if (input_record.line[t_open + 1] == '/')
      ret = TAG_CLOSE;
    else if (i > 0 && input_record.line[t_close - 1] == '/')
      ret = TAG_OPEN_CLOSE;
    else if (input_record.line[t_open + 1] == '?')
      ret = TAG_PROLOGUE;
    else if (input_record.line[t_open + 1] == '!')
      ret = TAG_DOCTYPE;
    else
      ret = TAG_OPEN;

    /* remove leading trim of all tags */
    if (t_open > 0)
    {
      lead_trim = 0;
      for (i = t_open - 1 ; i >= 0 && isspace(input_record.line[i]) ; i--)
        lead_trim++;
      if (output_record_length > lead_trim)
        output_record_length -= lead_trim;
      else
        output_record_length = 0;
    }

    /* update position */
    i_pos = t_close + 1;
  }

  return(ret);
}


void insert_delimiter(level)
int level;
{
  /* insert the delimiter string token between the fields */
  int i;
  if (level < nlevels && level > 0 && delimiter[level]->ndelimiters)
  {
    for (i = 0 ; i < delimiter[level]->length[0] ; i++)
      output_record[output_record_length++] = delimiter[level]->d[0][i];
  }
  else if (level > 1)
    output_record[output_record_length++] = ' ';
  else 
    output_record[output_record_length++] = '\n';
    /* perhaps ambiguous but also use new-line for level 0 */
}


void unescape_chars()
{
  /* reverts escaped XML chars back to their ASCII equivalent */
  long i, j;

  /* scan output_record and adjust length */
  for (i = 0 ; i < output_record_length ; i++)
  {
    if (output_record[i] == '&')
    {
      if (i + 3 < output_record_length &&
          output_record[i + 1] == 'l' &&
          output_record[i + 2] == 't' &&
          output_record[i + 3] == ';')
      {
        output_record[i] = '<';
        output_record_length -= 3;
        for (j = i + 1 ; j < output_record_length ; j++)
          output_record[j] = output_record[j + 3];
      }
      else if (i + 3 < output_record_length &&
          output_record[i + 1] == 'g' &&
          output_record[i + 2] == 't' &&
          output_record[i + 3] == ';')
      {
        output_record[i] = '>';
        output_record_length -= 3;
        for (j = i + 1 ; j < output_record_length ; j++)
          output_record[j] = output_record[j + 3];
      }
      else if (i + 4 < output_record_length &&
          output_record[i + 1] == 'a' &&
          output_record[i + 2] == 'm' &&
          output_record[i + 3] == 'p' &&
          output_record[i + 4] == ';')
      {
        output_record[i] = '&';
        output_record_length -= 4;
        for (j = i + 1 ; j < output_record_length ; j++)
          output_record[j] = output_record[j + 4];
      }
      else if (i + 5 < output_record_length &&
          output_record[i + 1] == 'q' &&
          output_record[i + 2] == 'u' &&
          output_record[i + 3] == 'o' &&
          output_record[i + 4] == 't' &&
          output_record[i + 5] == ';')
      {
        output_record[i] = '\"';
        output_record_length -= 5;
        for (j = i + 1 ; j < output_record_length ; j++)
          output_record[j] = output_record[j + 5];
      }
      else if (i + 7 < output_record_length &&
          output_record[i + 1] == 'b' &&
          output_record[i + 2] == 'r' &&
          output_record[i + 3] == 'v' &&
          output_record[i + 4] == 'b' &&
          output_record[i + 5] == 'a' &&
          output_record[i + 6] == 'r' &&
          output_record[i + 7] == ';')
      {
        output_record[i] = '|';
        output_record_length -= 7;
        for (j = i + 1 ; j < output_record_length ; j++)
          output_record[j] = output_record[j + 7];
      }
      else if (i + 5 < output_record_length &&
          output_record[i + 1] == 'a' &&
          output_record[i + 2] == 'p' &&
          output_record[i + 3] == 'o' &&
          output_record[i + 4] == 's' &&
          output_record[i + 5] == ';')
      {
        output_record[i] = '\'';
        output_record_length -= 5;
        for (j = i + 1 ; j < output_record_length ; j++)
          output_record[j] = output_record[j + 5];
      }
    }
  }
}
