/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2009, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert command line text output into XML formatted output     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "validate_args.h"

/* function templates */
void populate_delimiter(int level, unsigned char *d);
void increase_nlevels(int newlevel);
void allocate_attrname();
void version();
void usage(int invocation);

void validate_args(argc, argv, invocation)
int argc;
unsigned char **argv;
int invocation;
{
  /* validate all the parameters supplied to this command */
  int i, j, level, i_prev;
  int sw_printonly_firstinstance;

  sw_printonly_firstinstance = FALSE;

  /* only initialise once */
  if (invocation == ARGS_CMD)
  {
    /* local management switch for the printonly switches */
    sw_printonly_firstinstance = TRUE;

    /* initialise some infamous delimiter */
    delimiter_newline[0] = '\n'; delimiter_newline[1] = '\0';
    delimiter_cr[0]      = '\r'; delimiter_cr[1]      = '\0';
    delimiter_tab[0]     = '\t'; delimiter_tab[1]     = '\0';
    delimiter_space[0]   = ' ' ; delimiter_space[1]   = '\0';

    /* initialise default print switches */
    sw_print_default.tagname        = default_schema_field_name;
    sw_print_default.numbered_attr  = FALSE;
    sw_print_default.numbered_tag   = FALSE;
    sw_print_default.delimiter_attr = FALSE;
    sw_print_default.level_attr     = FALSE;
    sw_print_default.level_tag      = FALSE;
    sw_print_default.timestamp_attr = FALSE;

    /* initialise level delimiters and level switches */
    for (level = 0 ; level < nlevels ; level++)
    {
      delimiter[level]->ndelimiters = 0;

      if (level == 0)
      {
        sw_print[0]->tagname        = default_schema_parent_name;
        sw_print[0]->tagname_length = xstrlen(sw_print[0]->tagname);
        kvp_tag[0]->enabled         = FALSE;
        kvp_tag[0]->set             = FALSE;
        kvp_tag[0]->name            = sw_print[0]->tagname;
        kvp_tag[0]->length          = sw_print[0]->tagname_length;
        kvp_tag[0]->level_d         = 0;
        kvp_tag[0]->match_d         = DELIMITER_INDEX_UNMATCHED;
      }
      else if (level == 1)
      {
        delimiter[1]->match_d       = DELIMITER_INDEX_NEWLINE;
        sw_print[1]->tagname        = default_schema_line_name;
        sw_print[1]->tagname_length = xstrlen(sw_print[1]->tagname);
        kvp_tag[1]->enabled         = FALSE;
        kvp_tag[1]->set             = FALSE;
        kvp_tag[1]->name            = sw_print[1]->tagname;
        kvp_tag[1]->length          = sw_print[1]->tagname_length;
        kvp_tag[1]->level_d         = 0;
        kvp_tag[1]->match_d         = DELIMITER_INDEX_UNMATCHED;
      }
      else 
      {
        sw_print[level]->tagname        = sw_print_default.tagname;
        sw_print[level]->tagname_length = xstrlen(sw_print[level]->tagname);
        kvp_tag[level]->enabled         = FALSE;
        kvp_tag[level]->set             = FALSE;
        kvp_tag[level]->name            = sw_print[level]->tagname;
        kvp_tag[level]->length          = sw_print[level]->tagname_length;
        kvp_tag[level]->level_d         = 0;
        kvp_tag[level]->match_d         = DELIMITER_INDEX_UNMATCHED;
      }
      sw_print[level]->numbered_attr  = sw_print_default.numbered_attr;
      sw_print[level]->numbered_tag   = sw_print_default.numbered_tag;
      sw_print[level]->delimiter_attr = sw_print_default.delimiter_attr;
      sw_print[level]->level_attr     = sw_print_default.level_attr;
      sw_print[level]->level_tag      = sw_print_default.level_tag;
      sw_print[level]->timestamp_attr = sw_print_default.timestamp_attr;

      field_cntr[level] = 0;
    }

    /* initialise schema options */
    schema_argc = 0;

    /* initialise expel options */
    nexpel = 0;

    /* initialise column options */
    ncolumn = 0;

    /* initialise custom attribute options */
    nattrname = 0;

    /* initialise quotechar options */
    quotechars_length = 0;
    i_quotechar       = 0;
  }


  /* basic parameter handling  - deliberately avoiding getopt dependancy */
  i = 1;
  while (i < argc)
  {
    /* control */
    i_prev = i;


    /*
     * Option: Help
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--help") == 0 || strcmp((char *) argv[i], "-h") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        usage(invocation);
        i += 1;
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s cannot be specified in a schema file.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Version
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--version") == 0 || strcmp((char *) argv[i], "-v") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        version();
        i += 1;
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s cannot be specified in a schema file.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Debug
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--debug") == 0 || strcmp((char *) argv[i], "-debug") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        sw_debug = TRUE;
        i += 1;
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s cannot be specified in a schema file.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Linetags
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--linetags") == 0 || strcmp((char *) argv[i], "-L") == 0))
    {
      sw_line_tags = TRUE;
      i += 1;
    }


    /*
     * Option: Linenumbers
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--linenumbers") == 0 || strcmp((char *) argv[i], "-l") == 0))
    {
      sw_print[1]->numbered_tag = TRUE;
      i += 1;
    }


    /*
     * Option: Fieldnumbers
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--fieldnumbers") == 0 || strcmp((char *) argv[i], "-f") == 0))
    {
      sw_print[2]->numbered_tag = TRUE;
      i += 1;
    }


    /*
     * Option: Blanklines
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--blanklines") == 0 || strcmp((char *) argv[i], "-b") == 0))
    {
      sw_blank_lines = TRUE;
      i += 1;
    }


    /*
     * Option: Quoted fields
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--quotedfields") == 0 || strcmp((char *) argv[i], "-q") == 0))
    {
      sw_field_quoting = TRUE;
      i += 1;
    }


    /*
     * Option: Quoted fields spanning multiple input records
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--quotedfields2") == 0 || strcmp((char *) argv[i], "-q2") == 0))
    {
      sw_field_quoting_span  = TRUE;
      sw_field_quoting       = TRUE;
      i += 1;
    }


    /*
     * Option: Trim fields
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--trim") == 0 || strcmp((char *) argv[i], "-t") == 0))
    {
      sw_field_trim = TRUE;
      i += 1;
    }


    /*
     * Option: Expel empty fields
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--expelempty") == 0 || strcmp((char *) argv[i], "-e") == 0))
    {
      sw_field_expel_empty = TRUE;
      i += 1;
    }


    /*
     * Option: Print summary
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--summary") == 0 || strcmp((char *) argv[i], "-s") == 0))
    {
      sw_summary = TRUE;
      i += 1;
    }


    /*
     * Option: Print summary2
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--summary2") == 0 || strcmp((char *) argv[i], "-s2") == 0))
    {
      sw_summary = TRUE;
      sw_summary_stderr = TRUE;
      i += 1;
    }


    /*
     * Option: Undo xml
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--unxml") == 0 || strcmp((char *) argv[i], "-U") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        sw_xml_format = SW_UNXML;
        i += 1;
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s cannot be specified in a schema file.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: DTD schema file
     */
    if (i < argc
        && ((strcmp((char *) argv[i], "--schemadtd") == 0 || strcmp((char *) argv[i], "-Sd") == 0)
         || (strcmp((char *) argv[i], "--dtd") == 0 || strcmp((char *) argv[i], "-d") == 0)))
    {
      if (invocation != ARGS_SCHEMA)
      {
        if (sw_schema == SCHEMA_DEFAULT)
        {
          if (sw_key_value_pairs == FALSE)
          {
            if (i < argc - 1)
            {
              schema_filename = argv[i + 1];
              sw_schema = SCHEMA_DTD;
              i += 2;
            }
            else
            {
              fprintf(stderr, "%s: ERROR: Option %s requires a DTD schema file to be specified.\n", PROGNAME, argv[i]);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, "%s: ERROR: Option %s cannot be used with option --keyvaluepairs.\n", PROGNAME, argv[i]);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, "%s: ERROR: Option %s only one schema file can be specified.\n", PROGNAME, argv[i]);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s cannot be specified in a schema file.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: RNC schema file
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--schemarnc") == 0 || strcmp((char *) argv[i], "-Sr") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        if (sw_schema == SCHEMA_DEFAULT)
        {
          if (sw_key_value_pairs == FALSE)
          {
            if (i < argc - 1)
            {
              schema_filename = argv[i + 1];
              sw_schema = SCHEMA_RNC;
              i += 2;
            }
            else
            {
              fprintf(stderr, "%s: ERROR: Option %s requires a RNC schema file to be specified.\n", PROGNAME, argv[i]);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, "%s: ERROR: Option %s cannot be used with option --keyvaluepairs.\n", PROGNAME, argv[i]);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, "%s: ERROR: Option %s only one schema file can be specified.\n", PROGNAME, argv[i]);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s cannot be specified in a schema file.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: DTD or RNC schema file
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--schema") == 0 || strcmp((char *) argv[i], "-S") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        if (sw_schema == SCHEMA_DEFAULT)
        {
          if (sw_key_value_pairs == FALSE)
          {
            if (i < argc - 1)
            {
              j = xstrlen(argv[i + 1]);
              if (j > 4 && argv[i + 1][j - 4] == '.' && tolower(argv[i + 1][j - 3]) == 'd' && tolower(argv[i + 1][j - 2]) == 't' && tolower(argv[i + 1][j - 1]) == 'd')
                sw_schema = SCHEMA_DTD;
              else if (j > 4 && argv[i + 1][j - 4] == '.' && tolower(argv[i + 1][j - 3]) == 'r' && tolower(argv[i + 1][j - 2]) == 'n' && tolower(argv[i + 1][j - 1]) == 'c')
                sw_schema = SCHEMA_RNC;
              else
              {
                fprintf(stderr, "%s: ERROR: Option %s has an unsupported schema file extension %s.\n", PROGNAME, argv[i], argv[i + 1]);
                err_status = -1;
                usage(invocation);
              }
              schema_filename = argv[i + 1];
              i += 2;
            }
            else
            {
              fprintf(stderr, "%s: ERROR: Option %s requires a schema file to be specified.\n", PROGNAME, argv[i]);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, "%s: ERROR: Option %s cannot be used with option --keyvaluepairs.\n", PROGNAME, argv[i]);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, "%s: ERROR: Option %s only one schema file can be specified.\n", PROGNAME, argv[i]);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s cannot be specified in a schema file.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Record separator
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--recordseparator") == 0 || strcmp((char *) argv[i], "-R") == 0))
    {
      if (i < argc - 1)
      {
        populate_delimiter(1, argv[i + 1]);
        i += 2;
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s requires a delimiter to be specified.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Field separator
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--fieldseparator") == 0 || strcmp((char *) argv[i], "-F") == 0))
    {
      if (!ncolumn)
      {
        if (i < argc - 1)
        {
          populate_delimiter(2, argv[i + 1]);
          /* allow zero length fields */
          sw_field_empty_valid = TRUE;
          i += 2;
        }
        else
        {
          fprintf(stderr, "%s: ERROR: Option %s requires a delimiter to be specified.\n", PROGNAME, argv[i]);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s conflicts with option --column.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Delimiter
     */
    if (i < argc
        && ((strncmp((char *) argv[i], "--fieldseparator", 16) == 0 && xstrlen(argv[i]) > 16)
        || (strncmp((char *) argv[i], "-F", 2) == 0 && xstrlen(argv[i]) > 2)))
    {
      if (i < argc - 1)
      {
        level = process_arg_field(argv[i]);
        if (ncolumn == 0 || (ncolumn > 0 && level < 2))
        {
          if (level >= nlevels)
            increase_nlevels(level + 1);

          if (level > 0 && level < nlevels)
          {
            populate_delimiter(level, argv[i + 1]);
            if (level == 2) /* allow zero length fields */
              sw_field_empty_valid = TRUE;
            i += 2;
          }
          else 
          {
            err_status = -1;
            if (level == -15)
              fprintf(stderr, "%s: ERROR: Option %s has an invalid scope specified.\n", PROGNAME, argv[i]);
            else if (level == -13)
              fprintf(stderr, "%s: ERROR: Option %s has no scope specified.\n", PROGNAME, argv[i]);
            else if (level == -12)
              fprintf(stderr, "%s: ERROR: Option %s has no level specified.\n", PROGNAME, argv[i]);
            else
              fprintf(stderr, "%s: ERROR: Option %s has an invalid level specified.\n", PROGNAME, argv[i]);
            usage(invocation);
          }
        }
        else 
        {
          err_status = -1;
          fprintf(stderr, "%s: ERROR: Option %s conflicts with option --column.\n", PROGNAME, argv[i]);
          usage(invocation);
        }
      }
      else
      {
        err_status = -1;
        fprintf(stderr, "%s: ERROR: Option %s requires a delimiter to be specified.\n", PROGNAME, argv[i]);
        usage(invocation);
      }
    }


    /*
     * Option: Column mapping
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--column") == 0 || strcmp((char *) argv[i], "-C") == 0))
    {
      if (nlevels <= DEFAULT_NLEVELS && delimiter[2]->ndelimiters == 0)
      {
        if (i < argc - 1)
        {
          if (!process_arg_column(argv[i + 1]))
          {
            fprintf(stderr, "%s: ERROR: Option %s has an illegal range specified: %s\n", PROGNAME, argv[i], argv[i + 1]);
            err_status = -1;
            usage(invocation);
          }
          i += 2;
        }
        else
        {
          fprintf(stderr, "%s: ERROR: Option %s requires a range to be specified.\n", PROGNAME, argv[i]);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s cannot be used with delimiter levels of 2 or greater.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Tag name handling
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--tag") == 0 || strcmp((char *) argv[i], "-T") == 0))
    {
      if (i < argc - 1)
      {
        if (strcmp((char *) argv[i + 1], "number") == 0)
        {
          sw_print_default.numbered_tag = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->numbered_tag = sw_print_default.numbered_tag;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "level") == 0)
        {
          sw_print_default.level_tag = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->level_tag = sw_print_default.level_tag;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "tagname") == 0)
        {
          if (i < argc - 2)
          {
            sw_print_default.tagname = argv[i + 2];
            for (level = 1 ; level < nlevels ; level++)
            {
              sw_print[level]->tagname        = sw_print_default.tagname;
              sw_print[level]->tagname_length = xstrlen(sw_print[level]->tagname);
            }
            i += 3;
          }
          else
          {
            err_status = -1;
            fprintf(stderr, "%s: ERROR: Option %s %s requires a parameter to be specified.\n", PROGNAME, argv[i], argv[i + 1]);
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, "%s: ERROR: Option %s: Unsupported parameter: %s.\n", PROGNAME, argv[i], argv[i + 1]);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s requires a parameter to be specified.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }   
   
    if (i < argc
        && ((strncmp((char *) argv[i], "--tag", 5) == 0 && xstrlen(argv[i]) > 5)
        || (strncmp((char *) argv[i], "-T", 2) == 0 && xstrlen(argv[i]) > 2)))
    {
      if (i < argc - 1)
      {
        if (xstrlen(argv[i]) > 5)
        {
          level = atoi((char *) &argv[i][5]);
          /* atoi() is too smart */
          for (j = 5 ; j < xstrlen(argv[i]) ; j++)
            if (!isdigit(argv[i][j]))
              level = -1;
        }
        else
        {
          level = atoi((char *) &argv[i][2]);
          /* atoi() is too smart */
          for (j = 2 ; j < xstrlen(argv[i]) ; j++)
            if (!isdigit(argv[i][j]))
              level = -1;
        }
        if (level >= nlevels)
          increase_nlevels(level + 1);

        if (level >= 0 && level < nlevels)
        {
          if (strcmp((char *) argv[i + 1], "number") == 0)
          {
            sw_print[level]->numbered_tag = TRUE;
            i += 2;
          }
          else if (strcmp((char *) argv[i + 1], "level") == 0)
          {
            sw_print[level]->level_tag = TRUE;
            i += 2;
          }
          else if (strcmp((char *) argv[i + 1], "tagname") == 0)
          {
            if (i < argc - 2)
            {
              sw_print[level]->tagname        = argv[i + 2];
              sw_print[level]->tagname_length = xstrlen(sw_print[level]->tagname);
              i += 3;
            }
            else
            {
              err_status = -1;
              fprintf(stderr, "%s: ERROR: Option %s %s requires a parameter to be specified.\n", PROGNAME, argv[i], argv[i + 1]);
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, "%s: ERROR: Option %s: Unsupported parameter: %s.\n", PROGNAME, argv[i], argv[i + 1]);
            err_status = -1;
            usage(invocation);
          }
        }
        else 
        {
          err_status = -1;
          fprintf(stderr, "%s: ERROR: Option %s has an invalid level specified.\n", PROGNAME, argv[i]);
          usage(invocation);
        }
      }
      else
      {
        err_status = -1;
        fprintf(stderr, "%s: ERROR: Option %s requires a parameter to be specified.\n", PROGNAME, argv[i]);
        usage(invocation);
      }
    }

    
    /*
     * Option: Attribute handling
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--attribute") == 0 || strcmp((char *) argv[i], "-A") == 0))
    {
      if (i < argc - 1)
      {
        if (strcmp((char *) argv[i + 1], "number") == 0)
        {
          sw_print_default.numbered_attr = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->numbered_attr = sw_print_default.numbered_attr;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "delimiter") == 0)
        {
          sw_print_default.delimiter_attr = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->delimiter_attr = sw_print_default.delimiter_attr;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "level") == 0)
        {
          sw_print_default.level_attr = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->level_attr = sw_print_default.level_attr;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "timestamp") == 0)
        {
          sw_print_default.timestamp_attr = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->timestamp_attr = sw_print_default.timestamp_attr;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "attrname") == 0)
        {
          if (i < argc - 3)
          {
            allocate_attrname();
            attrname[nattrname].level        = -1; /* all levels */
            attrname[nattrname].name         = argv[i + 2];
            attrname[nattrname].name_length  = xstrlen(attrname[nattrname].name);
            attrname[nattrname].value        = argv[i + 3];
            attrname[nattrname].value_length = xstrlen(attrname[nattrname].value);
            nattrname++;
            i += 4;
          }
          else
          {
            fprintf(stderr, "%s: ERROR: Option %s %s requires 2 parameters to be specified.\n", PROGNAME, argv[i], argv[i + 1]);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, "%s: ERROR: Option %s: Unsupported parameter: %s.\n", PROGNAME, argv[i], argv[i + 1]);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s requires a parameter to be specified.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }

    if (i < argc
        && ((strncmp((char *) argv[i], "--attribute", 11) == 0 && xstrlen(argv[i]) > 11)
        || (strncmp((char *) argv[i], "-A", 2) == 0 && xstrlen(argv[i]) > 2)))
    {
      if (i < argc - 1)
      {
        if (xstrlen(argv[i]) > 11)
        {
          level = atoi((char *) &argv[i][11]);
          /* atoi() is too smart */
          for (j = 11 ; j < xstrlen(argv[i]) ; j++)
            if (!isdigit(argv[i][j]))
              level = -1;
        }
        else
        {
          level = atoi((char *) &argv[i][2]);
          /* atoi() is too smart */
          for (j = 2 ; j < xstrlen(argv[i]) ; j++)
            if (!isdigit(argv[i][j]))
              level = -1;
        }
        if (level >= nlevels)
          increase_nlevels(level + 1);

        if (level > 0 && level < nlevels)
        {
          if (strcmp((char *) argv[i + 1], "number") == 0)
          {
            sw_print[level]->numbered_attr = TRUE;
            i += 2;
          }
          else if (strcmp((char *) argv[i + 1], "delimiter") == 0)
          {
            sw_print[level]->delimiter_attr = TRUE;
            i += 2;
          }
          else if (strcmp((char *) argv[i + 1], "level") == 0)
          {
            sw_print[level]->level_attr = TRUE;
            i += 2;
          }
          else if (strcmp((char *) argv[i + 1], "timestamp") == 0)
          {
            sw_print[level]->timestamp_attr = TRUE;
            i += 2;
          }
          else if (strcmp((char *) argv[i + 1], "attrname") == 0)
          {
            if (i < argc - 3)
            {
              allocate_attrname();
              attrname[nattrname].level        = level;
              attrname[nattrname].name         = argv[i + 2];
              attrname[nattrname].name_length  = xstrlen(attrname[nattrname].name);
              attrname[nattrname].value        = argv[i + 3];
              attrname[nattrname].value_length = xstrlen(attrname[nattrname].value);
              nattrname++;
              i += 4;
            }
            else
            {
              fprintf(stderr, "%s: ERROR: Option %s %s requires 2 parameters to be specified.\n", PROGNAME, argv[i], argv[i + 1]);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, "%s: ERROR: Option %s: Unsupported parameter: %s.\n", PROGNAME, argv[i], argv[i + 1]);
            err_status = -1;
            usage(invocation);
          }
        }
        else 
        {
          err_status = -1;
          fprintf(stderr, "%s: ERROR: Option %s has an invalid level specified.\n", PROGNAME, argv[i]);
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s requires a parameter to be specified.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Print snippets
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--printonly") == 0 || strcmp((char *) argv[i], "-p") == 0))
    {
      /* set all switches to off but only for the first instance */
      if (sw_printonly_firstinstance)
      {
        sw_printonly_header        = FALSE;
        sw_printonly_footer        = FALSE;
        sw_printonly_rtagopen      = FALSE;
        sw_printonly_rtagclose     = FALSE;
        sw_printonly_records       = FALSE;
        sw_printonly_firstinstance = FALSE;
      }

      if (i < argc - 1)
      {
        if (strcmp((char *) argv[i + 1], "header") == 0)
        {
          sw_printonly_header = TRUE;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "footer") == 0)
        {
          sw_printonly_footer = TRUE;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "rtagopen") == 0)
        {
          sw_printonly_rtagopen = TRUE;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "rtagclose") == 0)
        {
          sw_printonly_rtagclose = TRUE;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "records") == 0)
        {
          sw_printonly_records = TRUE;
          i += 2;
        }
        else
        {
          fprintf(stderr, "%s: ERROR: Option %s: Unsupported parameter: %s.\n", PROGNAME, argv[i], argv[i + 1]);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s requires a parameter to be specified.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Key/value pairs
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--keyvaluepairs") == 0 || strcmp((char *) argv[i], "-k") == 0))
    {
      if (sw_schema == SCHEMA_DEFAULT)
      {
        sw_key_value_pairs = TRUE;
        sw_kvp_levels_all  = TRUE;
        i += 1;
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s cannot be used with a schema file.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }

    if (i < argc
        && ((strncmp((char *) argv[i], "--keyvaluepairs", 15) == 0 && xstrlen(argv[i]) > 15)
        || (strncmp((char *) argv[i], "-k", 2) == 0 && xstrlen(argv[i]) > 2)))
    {
      if (sw_schema == SCHEMA_DEFAULT)
      {
        if (xstrlen(argv[i]) > 15)
        {
          level = atoi((char *) &argv[i][15]);
          /* atoi() is too smart */
          for (j = 11 ; j < xstrlen(argv[i]) ; j++)
            if (!isdigit(argv[i][j]))
              level = -1;
        }
        else
        {
          level = atoi((char *) &argv[i][2]);
          /* atoi() is too smart */
          for (j = 2 ; j < xstrlen(argv[i]) ; j++)
            if (!isdigit(argv[i][j]))
              level = -1;
        }
        if (level >= nlevels)
          increase_nlevels(level + 1);
  
        if (level > 1 && level < nlevels)
        {
          sw_key_value_pairs      = TRUE;
          kvp_tag[level]->enabled = TRUE;
          i += 1;
        }
        else 
        {
          err_status = -1;
          fprintf(stderr, "%s: ERROR: Option %s has an invalid level specified.\n", PROGNAME, argv[i]);
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s cannot be used with a schema file.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Expel records and fields
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--expel") == 0 || strcmp((char *) argv[i], "-E") == 0))
    {
      if (i < argc - 1)
      {
        if (!process_arg_expel(argv[i + 1]))
        {
          fprintf(stderr, "%s: ERROR: Option %s has an illegal range specified: %s\n", PROGNAME, argv[i], argv[i + 1]);
          err_status = -1;
          usage(invocation);
        }
        i += 2;
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s requires a range to be specified.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Specify system identifier
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--identifier") == 0 || strcmp((char *) argv[i], "-I") == 0))
    {
      if (i < argc - 1)
      {
        schema_system_identifier = argv[i + 1];
        i += 2;
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s requires a system_identifier to be specified.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Quotechar
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--quotechars") == 0 || strcmp((char *) argv[i], "-Q") == 0))
    {
      if (!quotechars_length)
      {
        if (i < argc - 1)
        {
          quotechars = argv[i + 1];
          quotechars_length = xstrlen(quotechars);
          i += 2;
        }
        else
        {
          fprintf(stderr, "%s: ERROR: Option %s requires a quotechars array to be specified.\n", PROGNAME, argv[i]);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s only one quotechars array can be specified.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Specify XML handler
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--xmlformat") == 0 || strcmp((char *) argv[i], "-X") == 0))
    {
      if (i < argc - 1)
      {
        if (strcmp((char *) argv[i + 1], "XML1.0") == 0)
        {
          sw_xml_format = SW_XML;
          version_xml   = XML_1_0;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "XML1.1") == 0)
        {
          sw_xml_format = SW_XML;
          version_xml   = XML_1_1;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "SOAP1.1") == 0)
        {
          sw_xml_format = SW_SOAP;
          version_soap  = SOAP_1_1;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "SOAP1.2") == 0)
        {
          sw_xml_format = SW_SOAP;
          version_soap  = SOAP_1_2;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "BOM") == 0)
        {
          sw_bom = TRUE;
          i += 2;
        }
        else
        {
          fprintf(stderr, "%s: ERROR: Option %s: Unsupported parameter: %s.\n", PROGNAME, argv[i], argv[i + 1]);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, "%s: ERROR: Option %s requires a parameter to be specified.\n", PROGNAME, argv[i]);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Unmatched or unsupported option
     */
    if (i == i_prev)
    {
      fprintf(stderr, "%s: ERROR: Unknown option: %s\n", PROGNAME, argv[i]);
      err_status = -1;
      usage(invocation);
    }
  } /* end while */
}


void populate_delimiter(level, d)
int level;
unsigned char *d;
{
  /* add delimiter string for use at specified level */
  unsigned char *new_d;
  int i, j, new_d_length;

  /* assign delimiter */
  if (strcmp((char *) d, "\\n") == 0)       /* new line */
    new_d = delimiter_newline;
  else if (strcmp((char *) d, "\\r") == 0)  /* carriage return */
    new_d = delimiter_cr;
  else if (strcmp((char *) d, "\\t") == 0)  /* tab */
    new_d = delimiter_tab;
  else if (strcmp((char *) d, " ") == 0)    /* space */
    new_d = delimiter_space;
  else
    new_d = d;

  new_d_length = xstrlen(new_d);


  if (delimiter[level]->ndelimiters < 1)
  {
    delimiter[level]->d = (unsigned char **) malloc(sizeof(unsigned char *));
    if (delimiter[level]->d == NULL) out_of_mem();
    delimiter[level]->d[0] = (unsigned char *) malloc(sizeof(unsigned char));
    if (delimiter[level]->d[0] == NULL) out_of_mem();
    delimiter[level]->length = (int *) malloc(sizeof(int));
    if (delimiter[level]->length == NULL) out_of_mem();
    delimiter[level]->nscope = (int *) malloc(sizeof(int));
    if (delimiter[level]->nscope == NULL) out_of_mem();
    delimiter[level]->scope = (struct_scope **) malloc(sizeof(struct_scope *));
    if (delimiter[level]->scope == NULL) out_of_mem();
    delimiter[level]->scope_seq = (long *) malloc(sizeof(long));
    if (delimiter[level]->scope_seq == NULL) out_of_mem();
  }
  else
  {
    delimiter[level]->d = (unsigned char **) realloc(delimiter[level]->d, (delimiter[level]->ndelimiters + 1) * sizeof(unsigned char *));
    if (delimiter[level]->d == NULL) out_of_mem();
    delimiter[level]->d[delimiter[level]->ndelimiters] = (unsigned char *) malloc(sizeof(unsigned char));
    if (delimiter[level]->d[delimiter[level]->ndelimiters] == NULL) out_of_mem();
    delimiter[level]->length = (int *) realloc(delimiter[level]->length, (delimiter[level]->ndelimiters + 1) * sizeof(int));
    if (delimiter[level]->length == NULL) out_of_mem();
    delimiter[level]->nscope = (int *) realloc(delimiter[level]->nscope, (delimiter[level]->ndelimiters + 1) * sizeof(int));
    if (delimiter[level]->nscope == NULL) out_of_mem();
    delimiter[level]->scope = (struct_scope **) realloc(delimiter[level]->scope, (delimiter[level]->ndelimiters + 1) * sizeof(struct_scope *));
    if (delimiter[level]->scope == NULL) out_of_mem();
    delimiter[level]->scope_seq = (long *) realloc(delimiter[level]->scope_seq, (delimiter[level]->ndelimiters + 1) * sizeof(long));
    if (delimiter[level]->scope_seq == NULL) out_of_mem();
  }

  delimiter[level]->ndelimiters++;

  /* re-sort the delimiters in size from large to small */
  if (delimiter[level]->ndelimiters == 1)
  {
    delimiter[level]->d[0]         = new_d;
    delimiter[level]->length[0]    = new_d_length;
    delimiter[level]->nscope[0]    = nscope;
    delimiter[level]->scope[0]     = scope;
    delimiter[level]->scope_seq[0] = 0;
  }
  else
  {
    for (i = 0 ; i < delimiter[level]->ndelimiters - 1 ; i++)
    {
      if (new_d_length > delimiter[level]->length[i])
      {
        for (j = delimiter[level]->ndelimiters - 1 ; j > i ; j--)
        {
          delimiter[level]->d[j]      = delimiter[level]->d[j - 1];
          delimiter[level]->length[j] = delimiter[level]->length[j - 1];
          delimiter[level]->nscope[j] = delimiter[level]->nscope[j - 1];
          delimiter[level]->scope[j]  = delimiter[level]->scope[j - 1];
          /* delimiter[level]->scope_seq[j] is all initialised as 0 */
        }
        break; /* break for/i */
      }
    } /* end for/i */
    delimiter[level]->d[i]         = new_d;
    delimiter[level]->length[i]    = new_d_length;
    delimiter[level]->nscope[i]    = nscope;
    delimiter[level]->scope[i]     = scope;
    delimiter[level]->scope_seq[i] = 0;
  }
}


void increase_nlevels(newlevel)
int newlevel;
{
  /* allocate level objects for new level */

  /* allocate mem */
  delimiter = (struct_delimiter **) realloc(delimiter, newlevel * sizeof(struct_delimiter *));
  if (delimiter == NULL) out_of_mem();
  sw_print = (struct_sw_print **) realloc(sw_print, newlevel * sizeof(struct_sw_print *));
  if (sw_print == NULL) out_of_mem();
  kvp_tag = (struct_kvp_tag **) realloc(kvp_tag, newlevel * sizeof(struct_kvp_tag *));
  if (kvp_tag == NULL) out_of_mem();
  field_cntr = (int *) realloc(field_cntr, newlevel * sizeof(int));
  if (field_cntr == NULL) out_of_mem();

  /* update attributes of new levels */
  while (nlevels < newlevel)
  {
    delimiter[nlevels] = (struct_delimiter *) malloc(sizeof(struct_delimiter));
    delimiter[nlevels]->ndelimiters = 0;

    sw_print[nlevels] = (struct_sw_print *) malloc(sizeof(struct_sw_print));
    if (sw_print[nlevels] == NULL) out_of_mem();
    sw_print[nlevels]->tagname        = sw_print_default.tagname;
    sw_print[nlevels]->tagname_length = xstrlen(sw_print[nlevels]->tagname);
    sw_print[nlevels]->numbered_attr  = sw_print_default.numbered_attr;
    sw_print[nlevels]->numbered_tag   = sw_print_default.numbered_tag;
    sw_print[nlevels]->delimiter_attr = sw_print_default.delimiter_attr;
    sw_print[nlevels]->level_attr     = sw_print_default.level_attr;
    sw_print[nlevels]->level_tag      = sw_print_default.level_tag;
    sw_print[nlevels]->timestamp_attr = sw_print_default.timestamp_attr;

    kvp_tag[nlevels] = (struct_kvp_tag *) malloc(sizeof(struct_kvp_tag));
    if (kvp_tag[nlevels] == NULL) out_of_mem();
    kvp_tag[nlevels]->enabled = FALSE;
    kvp_tag[nlevels]->set     = FALSE;
    kvp_tag[nlevels]->name    = sw_print[nlevels]->tagname;
    kvp_tag[nlevels]->length  = sw_print[nlevels]->tagname_length;
    kvp_tag[nlevels]->level_d = 0;
    kvp_tag[nlevels]->match_d = DELIMITER_INDEX_UNMATCHED;

    field_cntr[nlevels] = 0;

    nlevels++;
  }
}


void allocate_attrname()
{
  if (nattrname)
  {
    attrname = (struct_attrname *) realloc(attrname, (nattrname + 1) * sizeof(struct_attrname));
    if (attrname == NULL) out_of_mem();
  }
  else
  {
    attrname = (struct_attrname *) malloc((nattrname + 1) * sizeof(struct_attrname));
    if (attrname == NULL) out_of_mem();
  }
}


void usage(invocation)
int invocation;
{
  /* Show quick usage */
  if (invocation == ARGS_SCHEMA)
    fprintf(stderr, "%s: ERROR: Illegal option is specified in schema file: %s\n", PROGNAME, schema_filename);

  fprintf(stderr, "Usage: %s [-h --help] [-v --version] [-debug --debug]\n", PROGNAME);
  fprintf(stderr, "             [-S --schema <file> | \\\n");
  fprintf(stderr, "                -Sd --schemadtd <file> | -Sr --schemarnc <file>]\n");
  fprintf(stderr, "             [-F --fieldseparator[<level>[:<scope>]] <string>]\n");
  fprintf(stderr, "             [-R --recordseparator <string>]\n");
  fprintf(stderr, "             [-C --column <c1>-<c2>[:<name>]]\n");
  fprintf(stderr, "             [-E --expel <input_records>[:<input_fields>]] [-e --expelempty]\n");
  fprintf(stderr, "             [-q --quotedfields[2]] [-Q --quotechars <string>]\n");
  fprintf(stderr, "             [-b --blanklines] [-t --trim] [-k --keyvaluepairs[<level>]]\n");
  fprintf(stderr, "             [-A --attribute[<level>] delimiter|number|level|timestamp\n");
  fprintf(stderr, "                |attrname <name> <value>]\n");
  fprintf(stderr, "             [-T --tag[<level>] number|level|tagname <name>]\n");
  fprintf(stderr, "             [-l --linenumbers] [-f --fieldnumbers] [-L --linetags]\n");
  fprintf(stderr, "             [-p --printonly header|footer|rtagopen|rtagclose|records]\n");
  fprintf(stderr, "             [-I --identifier <system_identifier>]\n");
  fprintf(stderr, "             [-X --xmlformat XML1.0|XML1.1|SOAP1.1|SOAP1.2|BOM]\n");
  fprintf(stderr, "             [-s --summary[2]] [-U --unxml]\n");
  exit(err_status);
}


void version()
{
  /* Show version */
  fprintf(stderr, "%s: Version %s\n", PROGNAME, VERSION);
  fprintf(stderr, "Copyright (c) 2008-2009, Arthur Gouros.\n");
  fprintf(stderr, "Refer to the %s license for terms and conditions.\n", PROGNAME);
  exit(err_status);
}

